# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import sys
import glob
from metagenomix._io_utils import caller, status_update, io_update, to_do


def get_sams_fastqs(
        mode: str,
        fastqs: dict
) -> dict:
    """Get the paths of the fastq files per sample or altogether.

    Parameters
    ----------
    mode : str
        Mode of processing for blobology:
        - "coassembly": all co-assembled samples.
        - "sample": each sample independently.
    fastqs : dict
        Path to the raw fastq files per sample.

    Returns
    -------
    sams_fastqs : dict
        Either same as `fastqs` of all fastqs under an empty key.
    """
    sams_fastqs = fastqs
    if mode == 'coassembly':
        sams_fastqs = {'': [y for x in fastqs.values() for y in x]}
    return sams_fastqs


def get_fqs(
        fastq_fps: list
) -> tuple:
    """Get the fastq files for the current sample to reassemble

    Parameters
    ----------
    fastq_fps : list
        Paths to fastq files for the current sample to reassemble

    Returns
    -------
    fqs : list
        Paths to gunzipped fastq files for the current sample to reassemble
    cmd : str
        gunzip commands
    """
    fqs = []
    cmd = ''
    for f in fastq_fps:
        if f.endswith('.gz'):
            f_q = f.replace('.gz', '')
            cmd += 'gunzip -c %s > %s\n' % (f, f_q)
        else:
            f_q = f
        if f_q.endswith('_R1.fastq'):
            f_q_ = f_q.replace('_R1.fastq', '_1.fastq')
            cmd += 'mv %s %s\n' % (f_q, f_q_)
        elif f_q.endswith('_R2.fastq'):
            f_q_ = f_q.replace('_R2.fastq', '_2.fastq')
            cmd += 'mv %s %s\n' % (f_q, f_q_)
        else:
            f_q_ = f_q
        fqs.append(f_q_)
    return fqs, cmd


def quantify_cmd(
        self,
        bins: str,
        fastq: list,
        out: str,
        contigs: str
) -> str:
    """Collect metaWRAP quantify command.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP
        .pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
        .status
            Tool status
    bins : str
        Path to the metaWRAP bins folder
    fastq : list
        Paths to the input fastq files
    out : str
        Paths to the output folder
    contigs : str
        Paths to the input contifs fasta file

    Returns
    -------
    cmd : str
        metaWRAP quantify command
    """
    fqs, fqs_cmd = get_fqs(fastq)
    cmd = 'metawrap quant_bins'
    cmd += ' -b %s' % bins
    cmd += ' -o %s' % out
    cmd += ' -a %s' % contigs
    cmd += ' -t %s' % self.soft.params['cpus']
    cmd += ' %s\n' % ' '.join(fqs)
    if fqs_cmd:
        cmd = fqs_cmd + cmd + 'rm %s\n' % ' '.join(fqs)
    return cmd


def get_fastqs(
        self,
        group: str,
) -> dict:
    """Collect the illumina reads for the samples of the current pool group.

    Parameters
    ----------
    self : Commands class instance
        .pool : str
            Pool name
        .pools : str
            All Pools and their samples
        .status
            Tool status
    group : str
        Current pool group

    Returns
    -------
    fastqs : dict
        Paths to fastqs files per sample
    """
    fastqs = {}
    for sam in self.pools[self.sam_pool][group]:
        if self.config.fastq_mv[sam].get(('illumina', sam)):
            fastqs[sam] = self.config.fastq_mv[sam][('illumina', sam)]
    return fastqs


def quantify(self):
    """Quantify the bins obtained using metaWRAP.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP
        .pool : str
            Pool name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
        .status
            Tool status
    """
    for (tech, group), inputs in self.inputs[self.sam_pool].items():
        key = (tech, group)
        self.outputs['outs'][key] = {}

        bins = inputs[0]
        assembler = self.softs['metawrap_binning'].prev
        contigs = self.softs[assembler].outputs[self.sam_pool][key][0]
        fastqs = get_fastqs(self, group)
        if not fastqs:
            self.soft.add_status(
                tech, self.sam_pool, group=group, message='no illumina data')
            print('[metawrap_blobology] No illumina reads for group %s' % group)
            continue

        io_update(self, i_f=contigs, i_d=bins, key=key)

        for mode in self.soft.params['blobology']:
            self.outputs['outs'][key][mode] = {}

            out_ = '%s/%s/%s/%s/%s' % (
                self.dir, tech, self.sam_pool, group, mode)
            sams_fastqs = get_sams_fastqs(mode, fastqs)
            for sam, fastq in sams_fastqs.items():
                to_dos = status_update(self, tech, fastq, group=group)
                to_dos.extend(status_update(
                    self, tech, [bins], group=group, folder=True))
                if sam:
                    out = out_ + '/%s' % sam
                else:
                    out = out_
                self.outputs['dirs'].append(out)

                cmd = quantify_cmd(self, bins, fastq, out, contigs)
                if self.config.force or not glob.glob('%s/*' % out):
                    self.outputs['outs'][key][mode][sam] = out
                    if to_dos:
                        self.outputs['cmds'].setdefault(key, []).append(False)
                    else:
                        self.outputs['cmds'].setdefault(key, []).append(cmd)
                    io_update(self, i_f=fastq, o_d=out, key=key)
                    self.soft.add_status(
                        tech, self.sam_pool, 1, group=group, genome=sam)
                else:
                    self.soft.add_status(
                        tech, self.sam_pool, 0, group=group, genome=sam)


def classify_or_annotate_dirs(
        self,
        inputs: list
) -> dict:
    """

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP
        .pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
        .status
            Tool status
    inputs : list
        Paths to the input files

    Returns
    -------
    bin_dirs : dict
        Path to the bins input folder per assembly solution
    """
    if self.soft.prev == 'metawrap_refine':
        bin_dirs = {'': [inputs[0]]}
    elif self.soft.prev == 'metawrap_reassemble':
        bin_dirs = inputs[self.sam_pool]
    else:
        sys.exit('No metawrap classify_bins after %s' % self.soft.prev)
    return bin_dirs


def classify_or_annotate_out(
        self,
        tech: str,
        bin_dir: str,
        group: str,
        sam: str
) -> str:
    """

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP
        .pool : str
            Pool name.
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    bin_dir : str
        Path to the bins input folder
    group : str
        Group for the current co-assembly and
    sam : str
        Name of a sample

    Returns
    -------
    out : str
        Path to the output folder
    """
    out = '/'.join([self.dir, tech, self.sam_pool, group])
    if 'permissive' in bin_dir:
        out += '/permissive'
    elif 'strict' in bin_dir:
        out += '/strict'
    if sam:
        out += '/%s' % sam
    return out


def classify_or_annotate_cmd(
        self,
        command: str,
        bin_dir: str,
        out: str
) -> str:
    """Collect metaWRAP classify_bins or annotate_bins command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    command : str
        "classify_bins" or "annotate_bins"
    bin_dir : str
        Path to the bins input folder
    out : str
        Path to the output folder

    Returns
    -------
    cmd : str
        metaWRAP classify_bins or annotate command_bins
    """
    cmd = 'metawrap %s' % command
    cmd += ' -b %s' % bin_dir
    cmd += ' -o %s' % out
    cmd += ' -t %s' % self.soft.params['cpus']
    return cmd


def classify_or_annotate(
        self,
        command: str
) -> None:
    """Classify or annotate the bins obtained using metaWRAP.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP
        .pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
        .status
            Tool status
    command : str
        "classify_bins" or "annotate_bins"
    """
    for (tech, group), inputs in self.inputs[self.sam_pool].items():
        key = (tech, group)
        self.outputs['outs'][key] = {}

        for sam, bins_dir in classify_or_annotate_dirs(self, inputs).items():
            for bin_dir in bins_dir:

                out = classify_or_annotate_out(self, tech, bin_dir, group, sam)
                self.outputs['dirs'].append(out)

                to_dos = status_update(
                    self, tech, [bin_dir], group=group, genome=sam)
                cmd = classify_or_annotate_cmd(self, command, bin_dir, out)
                self.outputs['outs'][key][sam] = [out]

                condition = classify_or_annotate_condition(command, out)
                if self.config.force or condition:
                    if to_dos:
                        self.outputs['cmds'].setdefault(key, []).append(False)
                    else:
                        self.outputs['cmds'].setdefault(key, []).append(cmd)
                    io_update(self, i_d=bin_dir, o_d=out, key=key)
                    self.soft.add_status(
                        tech, self.sam_pool, 1, group=group, genome=sam)
                else:
                    self.soft.add_status(
                        tech, self.sam_pool, 0, group=group, genome=sam)


def classify_or_annotate_condition(
        command: str,
        out: str
) -> bool:
    """Get the boolean telling whether the softwares was already run.

    Parameters
    ----------
    command : str
        "classify_bins" or "annotate_bins"
    out : str
        Path to the output folder

    Returns
    -------
    condition : bool
        True if classify_bins or annotate_bins was not already run
    """
    if command == 'annotate_bins':
        tab = '%s/bin_funct_annotations/*.tab' % out.replace(
            '${SCRATCH_FOLDER}', '')
        condition = not glob.glob(tab)
    elif command == 'classify_bins':
        tab = '%s/bin_taxonomy.tab' % out
        condition = to_do(tab)
    return condition


def classify(self):
    """Classify the bins obtained using metaWRAP.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP classify_bins
        .pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
        .status
            Tool status
    """
    classify_or_annotate(self, 'classify_bins')


def annotate(self):
    """Annotate the bins obtained using metaWRAP.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP annotate_bins
        .pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
        .status
            Tool status
    """
    classify_or_annotate(self, 'annotate_bins')


def blobology_cmd(
        self,
        fastq_fps: list,
        out: str,
        contigs: str,
        bins: str
) -> str:
    """Collect metaWRAP blobology command.

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
        .status
            Tool status
    fastq_fps : list
        Path the input fastq files
    out : str
        Path to the metaWRAP bin reassembly output folder
    contigs : str
        Path to the input contigs file
    bins : str
        Path to the input bins folder

    Returns
    -------
    cmd : str
        metaWRAP blobology command
    """
    fqs, fqs_cmd = get_fqs(fastq_fps)
    cmd = 'metawrap blobology'
    cmd += ' -o %s' % out
    cmd += ' -a %s' % contigs
    cmd += ' -t %s' % self.soft.params['cpus']
    cmd += ' --bins %s' % bins
    cmd += ' %s\n' % ' '.join(fqs)
    if fqs_cmd:
        cmd = fqs_cmd + cmd + 'rm %s\n' % ' '.join(fqs)
    return cmd


def blobology(self):
    """Make blobology plots for the bins obtained using metaWRAP.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP
        .pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
        .status
            Tool status
    """
    for (tech, group), inputs in self.inputs[self.sam_pool].items():
        key = (tech, group)

        self.outputs['outs'][key] = {}
        bins = inputs[0]

        assembler = self.softs['metawrap_binning'].prev
        contigs = self.softs[assembler].outputs[self.sam_pool][key][0]
        fastqs = get_fastqs(self, group)
        if not fastqs:
            # self.soft.add_status(tech, self.sam_pool, [fastqs], group=group)
            print('[metawrap_blobology] No illumina reads for group %s' % group)
            continue

        io_update(self, i_f=contigs, i_d=bins, key=key)
        for mode in self.soft.params['blobology']:
            self.outputs['outs'][key][mode] = {}
            out = '/'.join([self.dir, tech, self.sam_pool, group, mode])
            sams_fastqs = get_sams_fastqs(mode, fastqs)
            for sam, fastq_fps in sams_fastqs.items():
                to_dos = status_update(
                    self, tech, fastq_fps, group=group, genome=sam)
                to_dos.extend(status_update(self, tech, [bins], group=group,
                                            genome=sam, folder=True))
                if sam:
                    out += '/%s' % sam
                self.outputs['dirs'].append(out)
                plot = '%s/contigs.binned.blobplot' % out
                self.outputs['outs'][key][mode][sam] = out
                if self.config.force or to_do(plot):
                    cmd = blobology_cmd(self, fastq_fps, out, contigs, bins)
                    if to_dos:
                        self.outputs['cmds'].setdefault(key, []).append(False)
                    else:
                        self.outputs['cmds'].setdefault(key, []).append(cmd)
                    io_update(self, i_f=fastq_fps, o_d=out, key=key)
                    self.soft.add_status(
                        tech, self.sam_pool, 1, group=group, genome=sam)
                else:
                    self.soft.add_status(
                        tech, self.sam_pool, 0, group=group, genome=sam)


def reassembly_bins_cmd(
        self,
        fastq: list,
        out: str,
        bins: str
) -> str:
    """Collect metaWRAP bin reassembly command

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    fastq : list
        Paths to the input fastq files
    out : str
        Path to the metaWRAP bin reassembly output folder
    bins : str
        Path to the input bins folder

    Returns
    -------
    cmd : str
        metaWRAP bin reassembly command
    """
    fqs, fqs_cmd = get_fqs(fastq)
    cmd = 'metawrap reassemble_bins'
    cmd += ' -o %s' % out
    for idx, fq in enumerate(fqs):
        cmd += ' -%s %s' % ((idx + 1), fq)
    cmd += ' -t %s' % self.soft.params['cpus']
    cmd += ' -m %s' % self.soft.params['mem']
    cmd += ' -c %s' % self.soft.params['min_completion_reassembly']
    cmd += ' -x %s' % self.soft.params['min_contamination_reassembly']
    cmd += ' -b %s' % bins
    cmd += ' --skip-checkm'
    if self.soft.params['cpus'] > 1:
        cmd += ' --parallel'
    cmd += '\n'
    if fqs_cmd:
        cmd = fqs_cmd + cmd + 'rm %s\n' % ' '.join(fqs)
    return cmd


def reassembly_cmd(
        self,
        fastq: list,
        bins: str,
        tech: str,
        group: str,
        out: str,
        key: tuple
) -> str:
    """Collect metaWRAP bin reassembly command

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
        .status
            Tool status
    fastq : list
        Paths to the input fastq files
    bins : str
        Path to the input bins folder
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    group : str
        Name of the current co-assembly group
    out : str
        Path to the metaWRAP bin reassembly output folder
    key : tuple

    Returns
    -------
    cmd : str
        metaWRAP bin reassembly command
    """
    cmd = ''
    for mode in self.soft.params['reassembly']:
        mode_dir = '%s/reassembled_bins_%s' % (out, mode)
        self.outputs['dirs'].append(mode_dir)
        self.outputs['outs'].setdefault(key, []).append(mode_dir)
        io_update(self, i_d=mode_dir, key=key)
        if not self.config.force and not to_do(folder=mode_dir):
            continue
        cmd += 'mkdir %s\n' % mode_dir
        cmd += 'mv %s/reassembled_bins/*%s.fa %s/.\n' % (out, mode, mode_dir)
    if cmd:
        cmd = reassembly_bins_cmd(self, fastq, out, bins) + cmd
        self.soft.add_status(tech, self.sam_pool, 1, group=group)
    else:
        self.soft.add_status(tech, self.sam_pool, 0, group=group)
    return cmd


def reassemble(self):
    """Reassemble the bins obtained using metaWRAP.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP
        .pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
        .status
            Tool status
    """
    for (tech, group), inputs in self.inputs[self.sam_pool].items():
        key = (tech, group)

        bins = inputs[0]
        # if to_do(folder=bins):
        #     self.soft.add_status(tech, self.sam_pool, [bins], group=group)

        fastqs = get_fastqs(self, group)
        for sam, fastq in fastqs.items():
            if not fastq:
                print('[metawrap_reassemble] No illumina reads for %s' % sam)
                continue
            to_dos = status_update(self, tech, fastq, group=group, genome=sam)
            to_dos.extend(status_update(
                self, tech, [bins], group=group, genome=sam, folder=True))

            out = '/'.join([self.dir, tech, self.sam_pool, group, sam])
            self.outputs['dirs'].append(out)

            cmd = reassembly_cmd(self, fastq, bins, tech, group, out, key)
            if cmd:
                if to_dos:
                    self.outputs['cmds'].setdefault(key, []).append(False)
                else:
                    self.outputs['cmds'].setdefault(key, []).append(cmd)
                io_update(self, i_f=self.config.fastq[sam], i_d=bins,
                          o_d=out, key=key)
                self.soft.add_status(
                    tech, self.sam_pool, 1, group=group, genome=sam)
            else:
                self.soft.add_status(
                    tech, self.sam_pool, 0, group=group, genome=sam)


def refine_cmd(
        self,
        out_dir: str,
        bin_folders: list
) -> tuple:
    """Collect metaWRAP bin refinement command

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
        .config
            Configurations
    out_dir : str
        Path to the output folder
    bin_folders : list
        Path to the input bin folder

    Returns
    -------
    cmd : str
        metaWRAP bin refinement command
    n_bins : int
        Number of binners used
    """
    n_bins = 0
    cmd = 'metawrap bin_refinement'
    cmd += ' -o %s' % out_dir
    for fdx, folder in enumerate(bin_folders):
        folders = '%s/bin.*.fa' % folder.replace('${SCRATCH_FOLDER}', '')
        if len(glob.glob(folders)) or self.config.force:
            n_bins += 1
        cmd += ' -%s %s' % (['A', 'B', 'C'][fdx], folder)
    cmd += ' -t %s' % self.soft.params['cpus']
    cmd += ' -c %s' % self.soft.params['min_completion']
    cmd += ' -x %s' % self.soft.params['min_contamination']
    return cmd, n_bins


def refine(self):
    """Refine the bins obtained using metaWRAP.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP
        .pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    for (tech, group), inputs in self.inputs[self.sam_pool].items():
        key = (tech, group)

        out_dir = '/'.join([self.dir, tech, self.sam_pool, group])
        self.outputs['dirs'].append(out_dir)

        bin_folders = inputs[:-1]
        to_dos = status_update(
            self, tech, bin_folders, group=group, folder=True)

        cmd, n_bins = refine_cmd(self, out_dir, bin_folders)
        out = '%s/metawrap_%s_%s' % (out_dir,
                                     self.soft.params['min_completion'],
                                     self.soft.params['min_contamination'])
        stats, bins = '%s.stats' % out, '%s_bins' % out
        self.outputs['outs'][key] = [bins, stats]
        if not self.config.force and glob.glob('%s/*.fa' % bins):
            self.soft.add_status(tech, self.sam_pool, 0, group=group)
            continue
        if to_dos:
            self.outputs['cmds'].setdefault(key, []).append(False)
        else:
            self.outputs['cmds'].setdefault(key, []).append(cmd)
        io_update(self, i_d=bin_folders, o_f=stats, o_d=bins, key=key)
        self.soft.add_status(tech, self.sam_pool, 1, group=group)


def get_binners(
        self,
        binned: dict
) -> list:
    """Get the binning softwares that were not run yet.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
        .config
            Configurations
    binned : dict
        Path to the output folder per binner tool

    Returns
    -------
    binners : list
        Binner softwares used for the metaWRAP binning
    """
    binners = []
    for binner, bins in binned.items():
        bin_0 = '%s/bin.0.fa' % bins
        bin_1 = '%s/bin.1.fa' % bins
        if self.config.force or (to_do(bin_0) and to_do(bin_1)):
            binners.append(binner)
    return binners


def binning_cmd(
        self,
        fastq: list,
        out: str,
        contigs: str,
        binned: dict
) -> str:
    """Collect metaWRAP binning command

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    fastq : list
        Paths to the input fastq files
    out : str
        Path to the output folder
    contigs : str
        Path to the input contigs file
    binned : dict
        Path to the output folder per binner tool

    Returns
    -------
    cmd : str
        metaWRAP binning comand
    """
    binners = get_binners(self, binned)
    cmd = ''
    if binners:
        fqs, fqs_cmd = get_fqs(fastq)
        cmd += 'metawrap binning'
        cmd += ' -o %s' % out
        cmd += ' -a %s' % contigs
        cmd += ' -t %s' % self.soft.params['cpus']
        cmd += ' -l %s' % self.soft.params['min_contig_length']
        cmd += ' --universal'
        for binner in binners:
            cmd += ' --%s' % binner
        cmd += ' %s\n' % ' '.join(fqs)
        if fqs_cmd:
            cmd = fqs_cmd + cmd + 'rm %s\n' % ' '.join(fqs)
    return cmd


def binning(self):
    """Bin the assembled contigs using metaWRAP.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for metaWRAP
        .pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    for (tech, group), inputs in self.inputs[self.sam_pool].items():

        fastqs = [fastq for sam in self.pools[self.sam_pool][group] for fastq
                  in self.config.fastq_mv[sam].get(('illumina', sam), [])]

        to_dos = []
        if not fastqs:
            self.soft.add_status(tech, self.sam_pool, [fastqs], group=group,
                                 message='no illumina data')
            print('[metawrap_binning] No illumina reads for group %s' % group)
            continue
        else:
            to_dos.extend(status_update(self, tech, fastqs, group=group))

        contigs = inputs[0]
        to_dos.extend(status_update(self, tech, [contigs], group=group))

        tmp = '$TMPDIR/mtwrp_%s_%s_%s' % (self.sam_pool, tech, group)
        out = '/'.join([self.dir, tech, self.sam_pool, group])
        self.outputs['dirs'].append(out)

        binned = {binner: '%s/%s_bins' % (out, binner)
                  for binner in self.soft.params['binners']}
        bin_dirs = sorted(binned.values()) + ['%s/work_files' % out]

        key = (tech, group)
        self.outputs['outs'][key] = bin_dirs

        cmd = binning_cmd(self, fastqs, out, contigs, binned)
        if cmd:
            if to_dos:
                self.outputs['cmds'].setdefault(key, []).append(False)
            else:
                self.outputs['cmds'].setdefault(key, []).append(cmd)
            io_update(self, i_f=([contigs] + fastqs), i_d=tmp,
                      o_d=bin_dirs, key=key)
            self.soft.add_status(tech, self.sam_pool, 1, group=group)
        else:
            self.soft.add_status(tech, self.sam_pool, 0, group=group)


def metawrap(self) -> None:
    """MetaWRAP - a flexible pipeline for genome-resolved metagenomic data
    analysis.
    MetaWRAP aims to be an easy-to-use metagenomic wrapper suite that
    accomplishes the core tasks of metagenomic analysis from start to finish:
    read quality control, assembly, visualization, taxonomic profiling,
    extracting draft genomes (binning), and functional annotation. Additionally,
    metaWRAP takes bin extraction and analysis to the next level (see module
    overview below). While there is no single best approach for processing
    metagenomic data, metaWRAP is meant to be a fast and simple approach
    before you delve deeper into parameterization of your analysis. MetaWRAP
    can be applied to a variety of environments, including gut, water, and soil
    microbiomes (see metaWRAP paper for benchmarks). Each individual module of
    metaWRAP is a standalone program, which means you can use only the modules
    you are interested in for your data.

    References
    ----------
    Uritskiy, Gherman V., Jocelyne DiRuggiero, and James Taylor. "MetaWRAP—a
    flexible pipeline for genome-resolved metagenomic data analysis."
    Microbiome 6.1 (2018): 1-13.

    Notes
    -----
    GitHub  : https://github.com/bxlab/metaWRAP
    Paper   : https://doi.org/10.1186/s40168-018-0541-1

    Parameters
    ----------
    self : Commands class instance
        Contains all the attributes needed for binning on the current sample
    """
    # This function splits the name of the software and calls as function
    # the last underscore-separated field (which is in this module)
    module_call = caller(self, __name__)
    module_call(self)


def yamb(self):
    """
    YAMB (Yet Another Metagenome Binner) - semi-automatic pipeline for
    metagenomic contigs binning. It's based on t-Distributed Stochastic
    Neighbor Embedding (t-SNE) technique for dimensionality reduction,
    that uses tetranucleotide frequency distribution in metagenomic contigs
    and mean contig coverage, and HDBSCAN clustering method.

    References
    ----------
    Korzhenkov, A., 2019. YAMB: metagenome binning using nonlinear
    dimensionality reduction and density-based clustering. bioRxiv, p.521286.

    Notes
    -----
    GitHub  : https://github.com/laxeye/YAMB
    Paper   : https://doi.org/10.1101/521286

    Parameters
    ----------
    self : Commands class instance
        Contains all the attributes needed for binning on the current sample
    """
    pass


def summarize_yamb(self):
    pass


def mycc(self):
    """
    Automated binning tool that combines genomic signatures, marker genes and
    optional contig coverages within one or multiple samples, in order to
    visualize the metagenomes and to identify the reconstructed genomic
    fragments.

    References
    ----------
    Lin, H.H. and Liao, Y.C., 2016. Accurate binning of metagenomic
    contigs via automated clustering sequences using information of genomic
    signatures and marker genes. Scientific reports, 6(1), pp.1-8.

    Notes
    -----
    Code    : https://sourceforge.net/projects/sb2nhri/files/MyCC/
    Docs    : http://sourceforge.net/projects/sb2nhri/files/MyCC/.
    Paper   : https://doi.org/10.1038/srep24175

    Parameters
    ----------
    self : Commands class instance
        Contains all the attributes needed for binning on the current sample
    """
    pass



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import unittest
from metagenomix.core.pipeline import Graph, Soft


class TestGraph(unittest.TestCase):

    def test_add_edge(self):
        graph = Graph(3)
        pass

    def test_print_paths_util(self):
        graph = Graph(3)
        pass

    def test_print_paths(self):
        pass


class TestSoft(unittest.TestCase):

    def test_get_softs(self):

        soft = Soft()
        self.assertEqual(soft.prev, 'fastq')
        self.assertEqual(soft.name, '')

        soft.get_softs(['A', 'B'])
        self.assertEqual(soft.prev, 'A')
        self.assertEqual(soft.name, 'B')

        soft = Soft()
        soft.get_softs(['B'])
        self.assertEqual(soft.prev, 'fastq')
        self.assertEqual(soft.name, 'B')

        soft = Soft()
        soft.get_softs([[], []])
        self.assertEqual(soft.prev, [])
        self.assertEqual(soft.name, [])


class TestPipeline(unittest.TestCase):

    def setUp(self) -> None:
        # pipeline = Pipeline()
        pass

    def test_collect_soft_name(self):
        pass

    def test_get_names_idx(self):
        pass

    def test_make_graph(self):
        pass

    def test_get_paths(self):
        pass

    def test_init(self):
        pass

    def test_get_params(self):
        pass

    def test_generic_command(self):
        pass

    def test_get_input_path(self):
        pass

    def test_get_scratch(self):
        pass

    def test_run(self):
        pass


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import glob
from metagenomix._io_utils import caller, io_update, to_do, status_update
from metagenomix._inputs import (group_inputs, genome_key, genome_out_dir,
                                 get_extension, get_reads, add_folder)


def lorikeet_cmd(
        self,
        genome: str,
        fasta_folder: str,
        group_reads: dict,
        out_dir: str,
        key: tuple,
        step: str
) -> str:
    """Collect lorikeet call, lorikeet consensus, or lorikeet genotype command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    genome : str
        MAGs/Genomes folder name or empty string (for assembly contigs)
    fasta_folder : str
        Path to aseembly fasta or to an input folder with the genomes/MAGs
    group_reads : dict
        Path(s) to the fastq files per sample
    out_dir : str
        Path to the output folder
    key : tuple
        Turrent tech, pool, group, genome, etc
    step : str
        name of the Lorikeet module ("call", "consensus", or "genotype")

    Returns
    -------
    cmd : str
        lorikeet command
    """
    tmp_dir = '$TMPDIR/lorikeet_%s_%s' % ('_'.join(key), step)
    cmd = 'mkdir -p %s\n' % tmp_dir

    cmd += 'lorikeet %s' % step
    cmd += ' --output-directory %s' % out_dir
    cmd += ' --output-prefix %s_output' % step
    cmd += ' --bam-file-cache-directory %s' % tmp_dir
    cmd += ' --threads %s' % self.soft.params['cpus']
    if genome:
        cmd += ' --reference %s' % fasta_folder
    else:
        cmd += ' --genome-fasta-directory %s' % fasta_folder
        cmd += ' --genome-fasta-extension %s' % get_extension(self)

    cmd += ' --method %s' % ' '.join(self.soft.params['method'])
    cmd_single, cmd_coupled, cmd_longreads = '', '', ''
    for sam, tech_sam_fastqs in group_reads.items():
        for tech_sam, fastqs in tech_sam_fastqs.items():
            if tech_sam[0] == 'illumina':
                if len(fastqs) == 1:
                    cmd_single += ' %s' % fastqs[0]
                elif len(fastqs) == 2:
                    cmd_coupled += ' %s' % ' '.join(fastqs)
            elif fastqs:
                cmd_longreads += ' %s' % fastqs[0]

    if cmd_single:
        cmd += ' --single%s' % cmd_single
    if cmd_coupled:
        cmd += ' --coupled%s' % cmd_coupled
    if cmd_longreads:
        cmd += ' --longreads%s' % cmd_longreads

    for param in [
        'min_read_aligned_length', 'min_read_percent_identity',
        'min_read_aligned_percent', 'min_read_aligned_length_pair',
        'min_read_percent_identity_pair', 'min_read_aligned_percent_pair',
        'min_covered_fraction', 'contig_end_exclusion', 'trim_min', 'trim_max',
        'mapper', 'longread_mapper', 'kmer_sizes', 'ploidy',
        'qual_by_depth_filter', 'qual_threshold', 'depth_per_sample_filter',
        'min_base_quality', 'min_mapq', 'base_quality_score_threshold',
        'max_input_depth', 'min_contig_size', 'min_sv_qual',
        'phred_scaled_global_read_mismapping_rate',
        'pair_hmm_gap_continuation_penalty',
        'heterozygosity', 'heterozygosity_stdev', 'indel_heterozygosity',
        'standard_min_confidence_threshold_for_calling',
        'active_probability_threshold', 'min_assembly_region_size',
        'max_assembly_region_size', 'assembly_region_padding',
        'min_dangling_branch_length', 'min_prune_factor', 'num_pruning_samples',
        'initial_error_rate_for_pruning', 'pruning_log_odds_threshold',
        'max_unpruned_variants', 'max_prob_propagation_distance',
        'max_mnp_distance', 'parallel_genomes'
    ]:
        cmd += ' --%s %s' % (param.replace('_', '-'), self.soft.params[param])
    for boolean in [
        'discard_improper_pairs', 'discard_supplementary', 'include_secondary',
        'discard_unmapped', 'high_memory', 'splitbams',
        'minimap2_reference_is_index', 'calculate_fst', 'calculate_dnds',
        'features_vcf', 'do_not_call_svs', 'use_posteriors_to_calculate_qual',
        'annotate_with_num_discovered_alleles',
        'dont_increase_kmer_sizes_for_cycles', 'allow_non_unique_kmers_in_ref',
        'do_not_run_physical_phasing', 'recover_all_dangling_branches',
        'use_adaptive_pruning', 'graph_output', 'dont_use_soft_clipped_bases',
        'disable_optimizations', 'disable_avx'
    ]:
        if self.soft.params[boolean]:
            cmd += ' --%s' % boolean.replace('_', '-')
    for param in ['minimap2_params', 'bwa_params', 'ngmlr_params']:
        if self.soft.params[param]:
            cmd += ' --%s "%s"' % (
                param.replace('_', '-'), ' '.join(self.soft.params[param]))
    cmd += ' --force'
    return cmd


def call(
        self,
        tech: str,
        fastas_folders: dict,
        reads: dict,
        group: str
) -> None:
    """Call variants using local reassembly across multiple genomes and samples.

    Parameters
    ----------
    self : Commands class instance
        .soft.prev : str
            Previous software in the pipeline
        .dir : str
            Path to pipeline output folder for lorikeet call
        .sam_pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    tech : str
        Technology or technologies used for assembly
    fastas_folders : dict
        Path(s) to aseembly fasta or to the input folders with the genomes/MAGs
    reads : dict
        Path(s) to the input fastq file(s) per sample and tech/sample
    group : str
        Name of a co-assembly pool's group
    """
    get_lorikeet(self, tech, fastas_folders, reads, group, 'call')


def consensus(
        self,
        tech: str,
        fastas_folders: dict,
        reads: dict,
        group: str
) -> None:
    """Generate consensus genomes for each provided sample and genome.

    Parameters
    ----------
    self : Commands class instance
        .soft.prev : str
            Previous software in the pipeline
        .dir : str
            Path to pipeline output folder for lorikeet consensus
        .sam_pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    tech : str
        Technology or technologies used for assembly
    fastas_folders : dict
        Path(s) to aseembly fasta or to the input folders with the genomes/MAGs
    reads : dict
        Path(s) to the input fastq file(s) per sample and tech/sample
    group : str
        Name of a co-assembly pool's group
    """
    get_lorikeet(self, tech, fastas_folders, reads, group, 'consensus')


def genotype(
        self,
        tech: str,
        fastas_folders: dict,
        reads: dict,
        group: str
) -> None:
    """Resolves strain-level genotypes and abundance from metagenomes.

    Parameters
    ----------
    self : Commands class instance
        .soft.prev : str
            Previous software in the pipeline
        .dir : str
            Path to pipeline output folder for lorikeet genotype
        .sam_pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    tech : str
        Technology or technologies used for assembly
    fastas_folders : dict
        Path(s) to aseembly fasta or to the input folders with the genomes/MAGs
    reads : dict
        Path(s) to the input fastq file(s) per sample and tech/sample
    group : str
        Name of a co-assembly pool's group
    """
    get_lorikeet(self, tech, fastas_folders, reads, group, 'genotype')


def get_lorikeet(
        self,
        tech: str,
        fastas_folders: dict,
        reads: dict,
        group: str,
        step: str
) -> None:
    """Get the Lorikeet inputs and outputs for the current tech and
    MAGs/genomes and collect the commands and data structures.

    Parameters
    ----------
    self : Commands class instance
        .soft.prev : str
            Previous software in the pipeline
        .dir : str
            Path to pipeline output folder for checkm tetra
        .sam_pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    tech : str
        Technology or technologies used for assembly
    fastas_folders : dict
        Path(s) to aseembly fasta or to the input folders with the genomes/MAGs
    reads : dict
        Path(s) to the input fastq file(s) per sample and tech/sample
    group : str
        Name of a co-assembly pool's group
    step : str
        Name of the Lorikeet step
    """
    for genome, fasta_folder_ in fastas_folders.items():

        fasta_folder = fasta_folder_[0]
        out_dir = genome_out_dir(self, tech, fasta_folder, group, genome)
        if self.soft.name == 'lorikeet':
            out_dir = add_folder(self, 'lorikeet', out_dir, step)
        self.outputs['dirs'].append(out_dir)
        self.outputs['outs'].setdefault((tech, group), []).append(out_dir)

        key = genome_key(tech, group, genome)
        if genome:
            to_dos = status_update(self, tech, [fasta_folder], folder=True)
        else:
            to_dos = status_update(self, tech, [fasta_folder])

        if self.config.force or not glob.glob('%s/*' % out_dir):
            if self.sam_pool:
                groups = self.pools[self.sam_pool][group]
                group_reads = dict(x for x in reads.items() if x[0] in groups)
            else:
                group_reads = reads

            cmd = lorikeet_cmd(
                self, genome, fasta_folder, group_reads, out_dir, key, step)
            if to_dos:
                self.outputs['cmds'].setdefault(key, []).append(False)
            else:
                self.outputs['cmds'].setdefault(key, []).append(cmd)
            if genome:
                io_update(self, i_d=fasta_folder, o_d=out_dir, key=key)
            else:
                io_update(self, i_f=fasta_folder, o_d=out_dir, key=key)


def lorikeet_(
        self,
        tech: str,
        fastas_folders: dict,
        reads: dict,
        group: str
) -> None:
    """Perform all the key steps of the lorikeet analysis.

    Notes
    -----
    This is done if the pipeline specifies "drep lorikeet" and not specific
    modules of checkm, such as "drep lorikeet_call" or "drep lorikeet_evolve"

    Parameters
    ----------
    self : Commands class instance
        .soft.prev : str
            Previous software in the pipeline
        .dir : str
            Path to pipeline output folder for checkm tetra
        .sam_pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    tech : str
        Technology or technologies used for assembly
    fastas_folders : dict
        Path(s) to aseembly fasta or to the input folders with the genomes/MAGs
    reads : dict
        Path(s) to the input fastq file(s) per sample and tech/sample
    group : str
        Name of a co-assembly pool's group
    """
    for step in ['call', 'consensus', 'genotype']:
        get_lorikeet(self, tech, fastas_folders, reads, group, step)


def lorikeet(self) -> None:
    """Lorikeet aims to call variants in metagenomes using local reassembly
    of haplotypes. Within-species variant analysis pipeline for metagenomic
    communities that utilizes both long and short reads. Lorikeet utilizes a
    re-implementaion of the GATK HaplotypeCaller algorithm, performing local
    re-assembly of potentially active regions within candidate genomes.
    Called variants can be clustered into likely strains using a combination
    of UMAP and HDBSCAN. Additional statistics, like consensus ANI,
    population ANI, and subpopulation ANI will also be calculated for each
    input geome providing values for each sample compared to the reference
    and also compared to all other samples.

    Lorikeet has a variety of subcommands with the main being call and
    genotype. The call pipeline will take any number of input genomes and
    samples and perform robust variant calling and ANI calculations. The
    genotype algorithm takes this a step further and attempts to reconstruct
    strain haplotypes from the called variants and return complete strain
    genomes.

    Input can either be reads and reference genome, or MAG. Or a BAM file
    and associated genome.

    References
    ----------
    In prep

    Notes
    -----
    GitHub  : https://github.com/rhysnewell/Lorikeet
    Docs    : https://rhysnewell.github.io/Lorikeet/
    Paper   : In prep

    Parameters
    ----------
    self : Commands class instance
        .soft.name : str
            Current software in the pipeline
        .soft.prev : str
            Previous software in the pipeline
        .dir : str
            Path to pipeline output folder for checkm tetra
        .sam_pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    reads = get_reads(self)

    __module_call__ = caller(self, __name__)
    if self.sam_pool in self.pools:
        for (tech, group), inputs in self.inputs[self.sam_pool].items():
            fastas_folders = group_inputs(self, inputs, True)
            __module_call__(self, tech, fastas_folders, reads, group)

    elif set(self.inputs) == {''}:
        for (tech, bin_algo), inputs in self.inputs[''].items():
            folders = group_inputs(self, inputs, True)
            __module_call__(self, tech, folders, reads, bin_algo)


def instrain(self):
    pass



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import yaml
import glob
import shutil
import random
import subprocess
import pkg_resources
import numpy as np
import datetime as dt
from tabulate import tabulate
from os.path import dirname, exists, isdir, islink, splitext

from metagenomix._io_utils import (
    mkdr, get_roundtrip, print_status_table, compute_hash, get_md5, get_dates)

STORAGE = pkg_resources.resource_filename("metagenomix", "storage")


class Created(object):

    def __init__(self, config, databases, workflow, commands):
        self.config = config
        self.databases = databases
        self.commands = commands
        self.graph = workflow.graph
        self.cmd = []
        self.cmds = {}
        self.links = {}
        self.soft_links = {}
        self.links_stats = {}
        self.avail = {}
        self.chunks = {}
        self.soft = ''
        self.hash = ''
        self.sh = ''
        self.main_sh = ''
        self.move_sh = ''
        self.move_shs = {}
        self.run = {'database': {}, 'software': {}}
        self.job_fps = []
        self.job_name = ''
        self.modules = {}
        self.pjct = self.get_prjct()
        self.scheduler = self.get_scheduler()
        self.time = dt.datetime.now().strftime("%d/%m/%Y-%H:%M")
        self.scripts = []
        self.scripts_parts = []
        self.log_dir = '%s/_created' % config.dir
        self.link_script = None

    def make_dirs(self):
        for name, soft in self.commands.softs.items():
            for directory in sorted(soft.dirs):
                if not islink(directory) and not isdir(directory):
                    os.makedirs(directory)

    def get_scheduler(self):
        if self.config.jobs:
            if self.config.torque:
                return 'qsub'
            else:
                return 'sbatch'
        else:
            return 'sh'

    def get_prjct(self):
        prjct = [x for x in self.config.project if x.lower() not in 'aeiouy']
        if prjct:
            return ''.join(prjct)
        else:
            return self.config.project

    def get_chunks(self, n_chunks):
        self.chunks = {}
        if n_chunks and len(self.cmds) >= n_chunks:
            cmds_d = dict(enumerate(self.cmds))
            chunks = [list(x) for x in np.array_split(list(cmds_d), n_chunks)]
            for cdx, chunk in enumerate(chunks):
                if len(chunk):
                    self.chunks[str(cdx)] = [cmds_d[x] for x in chunk]
        else:
            for key in self.cmds:
                chunkey = '%s_%s' % (key[0], '-'.join(key[1]).replace('/', '_'))
                self.chunks[chunkey] = [key]

    def get_main_sh(self, name, soft=None, local='') -> None:
        if soft:
            n = 'software'
            key = '%s%s' % (name, local)
            if self.cmds:
                self.main_sh = '%s/run%s.sh' % (soft.dir, local)
                self.run[n][key] = self.main_sh
            else:
                self.run[n][key] = 'Input needed (output from %s)' % soft.prev
        else:
            self.main_sh = '%s/run_%s.sh' % (self.sh.rsplit('/', 2)[0], name)
            self.run['database'][name] = self.main_sh

    def write_main(self) -> None:
        with open(self.main_sh, 'w') as o:
            if len(self.job_fps):
                job_dir = dirname(self.job_fps[0])
                o.write('mkdir -p %s/output\n' % job_dir)
                o.write('cd %s/output\n' % job_dir)
                for jdx, job_fp in enumerate(self.job_fps):
                    o.write('%s %s\n' % (self.scheduler, job_fp))
                self.job_fps = []

    def write_bash(self, name, soft) -> None:
        if soft.bash:
            self.get_main_sh(name, soft, '_locally')
            with open(self.main_sh, 'w') as o:
                for cmd in soft.bash:
                    o.write('%s\n' % cmd)

    def get_provenance_fp(self, name, soft=None) -> str:
        if soft:
            fp = '%s/provenance.txt' % soft.dir
        else:
            fp = '%s/provenance_%s.txt' % (self.sh.rsplit('/', 2)[0], name)
        return fp

    def write_provenance(self, name, soft=None) -> None:
        steps = self.graph.paths[soft.name][0]
        step_max = max([len(x) for x in steps])
        provenance_fp = self.get_provenance_fp(name, soft)
        with open(provenance_fp, 'w') as o:
            o.write("Pipeline steps to this output (and analysis type):\n")
            for sdx, step in enumerate(steps):
                o.write('%s\t%s%s\t: %s\n' % (
                    sdx, step, (' ' * (step_max-len(step))),
                    self.config.tools.get(step, '')))
            o.write("\nParameters for these steps' outputs:\n")
            for sdx, step in enumerate(steps):
                if step not in self.commands.softs:
                    o.write('\n ===== %s: %s (no param) =====\n' % (sdx, step))
                    continue
                o.write('\n========== %s: %s ==========\n' % (sdx, step))
                params_show = self.commands.softs[step].params.copy()
                if 'databases' in params_show:
                    databases = params_show['databases']
                    del params_show['databases']
                    params = {'search': {soft.name.split('_')[-1]: params_show,
                                         'databases': databases}}
                    o.write('%s\n' % yaml.dump(params))
                else:
                    o.write('%s\n' % yaml.dump(params_show))

    def database_cmds(self):
        for db, cmds in self.databases.commands.items():
            self.cmds = cmds
            self.get_chunks(self.config.params['chunks'])
            self.write_jobs(db)
            self.write_main(db)

    def get_modules(self, name: str):
        self.modules = self.config.modules.get(name, [])

    def scratch(self, soft, key, cmds):
        if soft.params['scratch'] and self.config.jobs and key in soft.io:
            roundtrip = get_roundtrip(soft.io[key])
            extended_cmds = ['\n# Move to SCRATCH_FOLDER']
            extended_cmds += roundtrip['to']
            extended_cmds += ['\n# data: %s %s' % (key[0], ' '.join(key[1]))]
            extended_cmds += cmds
            if self.config.move_back:
                extended_cmds += ['\n# Move from SCRATCH_FOLDER']
                extended_cmds += roundtrip['from']
        else:
            extended_cmds = ['\n# data: %s %s' % (key[0], ' '.join(key[1]))]
            extended_cmds += cmds
        self.cmds[key] = extended_cmds
        self.dummy_outputs(soft, key)

    def dummy_outputs(self, soft, key):
        if self.config.dev:

            for fil_ in soft.io[key].get(('O', 'f'), []):
                if fil_.endswith('*'):
                    continue
                fil_loc = fil_.replace('${SCRATCH_FOLDER}', '')
                fil_sto = fil_loc.replace(dirname(self.config.dir), STORAGE)
                if exists(fil_loc):
                    os.remove(fil_loc)
                if exists(fil_sto):
                    os.remove(fil_sto)

            for fol_ in soft.io[key].get(('O', 'd'), []):
                if fol_.endswith('*'):
                    continue
                fol_loc = fol_.replace('${SCRATCH_FOLDER}', '')
                fol_sto = fol_loc.replace(dirname(self.config.dir), STORAGE)
                if islink(fol_loc):
                    os.unlink(fol_loc)
                elif isdir(fol_loc):
                    shutil.rmtree(fol_loc)
                if isdir(fol_sto):
                    shutil.rmtree(fol_sto)

            if random.choice([0, 1]):

                for fil_ in soft.io[key].get(('O', 'f'), []):
                    if fil_.endswith('*'):
                        continue
                    fil_loc = fil_.replace('${SCRATCH_FOLDER}', '')
                    fil_sto = fil_loc.replace(dirname(self.config.dir), STORAGE)
                    if not isdir(dirname(fil_loc)):
                        os.makedirs(dirname(fil_loc))
                    if not isdir(dirname(fil_sto)):
                        os.makedirs(dirname(fil_sto))
                    with open(fil_sto, 'w') as o:
                        for r in range(10, random.randrange(11, 100)):
                            o.write('test\n')
                    os.symlink(fil_sto, fil_loc)

                for fol_ in soft.io[key].get(('O', 'd'), []):
                    if fol_.endswith('*'):
                        continue
                    fol_loc = fol_.replace('${SCRATCH_FOLDER}', '')
                    fol_sto = fol_loc.replace(dirname(self.config.dir), STORAGE)
                    if not isdir(fol_sto):
                        os.makedirs(fol_sto)
                    if not isdir(dirname(fol_loc)):
                        os.makedirs(dirname(fol_loc))
                    if not isdir(fol_loc):
                        os.symlink(fol_sto, fol_loc)

    def get_cmds(self, soft):
        self.cmds = {}
        for sam_or_pool, cmds in soft.cmds.items():
            if sum(map(bool, cmds)) or soft.name == 'pooling':
                self.scratch(soft, sam_or_pool, cmds)
                self.avail.setdefault(soft.name, []).append(sam_or_pool)

    def print_status(self, m, sdx, name, soft):
        gap = (m - len(name) - len(str(sdx))) + 1
        print('\t%s [%s] %s%s' % (sdx, name, ('.' * gap), ('.' * 8)), end=' ')
        print_status_table(soft, self.config.show_status)

    def write_logs(self):
        if not isdir(self.log_dir):
            os.makedirs(self.log_dir)
        fp = '%s/%s.txt' % (self.log_dir, compute_hash(self.hash))
        logs = self.get_logs(get_dates(fp))
        with open(fp, 'w') as o:
            for log in logs:
                o.write('%s\n' % log)
            if self.link_script:
                o.write(self.link_script)
        print('\n[config] Written: %s' % fp)

    def get_logs(self, dates):
        logs = self.scripts
        logs.extend(self.get_date_logs(dates))
        logs.extend(self.get_input_logs())
        return logs

    def get_date_logs(self, dates) -> list:
        date_logs = ['\n------------------']
        date_logs.extend(['Date: %s' % date for date in dates])
        date_logs.append('Date: %s' % self.time)
        date_logs.append('------------------')
        return date_logs

    def get_input_logs(self):
        input_logs = ['\n* Input files:']
        for key, arg in [
            ('illumina_dirs', '--fastq-dir-illumina'),
            ('pacbio_dirs', '--fastq-dir-pacbio'),
            ('nanopore_dirs', '--fastq-dir-nanopore'),
            ('output_dir', '--output-dir'),
            ('meta_fp', '--metadata'),
            ('pipeline_tsv', '--pipeline'),
            ('databases_yml', '--databases'),
            ('user_params_yml', '--user-params'),
            ('coassembly_yml', '--co-assembly'),
            ('strains_yml', '--strains')
        ]:
            vals = self.config.__dict__[key]
            if not vals:
                continue
            if not isinstance(vals, tuple):
                vals = (vals,)
            for val in vals:
                if key == 'meta_fp':
                    input_logs.append('* Input configuration files:')
                input_logs.append('%s%s%s\t%s' % (arg, ' ' * (25-len(arg)),
                                                  get_md5(val), val))
        return input_logs

    def get_links_chunks(self):
        cmds = {}
        if self.config.links_chunks:
            n_chunks = self.config.links_chunks
            links = set([y for x in self.soft_links.values() for y in x])
            if n_chunks and 1 < n_chunks <= len(links):
                cmds_ = dict(enumerate(list(links)))
                chunk = [list(x) for x in np.array_split(list(cmds_), n_chunks)]
                for cdx, c in enumerate(chunk):
                    if len(c):
                        cmds['-%s' % str(cdx + 1)] = [cmds_[x] for x in c]
            elif links:
                cmds[''] = list(links)
        else:
            for name, links in self.soft_links.items():
                if links:
                    cmds['-%s' % name] = links
        return cmds

    def get_links_dir(self):
        links_dir = '%s/%s' % (self.log_dir, compute_hash(self.hash))
        if not isdir(links_dir):
            os.makedirs(links_dir)
        if not isdir('%s/scripts' % links_dir):
            os.makedirs('%s/scripts' % links_dir)
        return links_dir

    def write_screen_jobs(self, links_dir, scripts):
        self.move_sh = '%s/move.sh' % links_dir
        with open(self.move_sh, 'w') as o:
            for name, (screen_sh, out) in sorted(scripts.items()):
                o.write('sh %s\n' % screen_sh)
            o.write('screen -ls\n')
            o.write('echo "To list running screen session(s): screen -ls"\n')
            o.write('echo "To get into a screen session: screen -r <ID>"\n')
            o.write('echo "To detach when within screen session: <ctrl-d>"\n')
            o.write('echo "To kill a screen session from within: <ctrl-k>"\n')

    def write_links(self):
        links_dir = self.get_links_dir()
        scripts = self.get_bring_links_scripts(links_dir)
        if scripts:
            self.write_screen_jobs(links_dir, scripts)
            self.print_links(scripts)

    def print_links(self, scripts):
        s = '\n%s\n' % ('*' * 46)
        m = '%sSome data needed for these jobs is stored away%s' % (s, s)
        tab = [x for x in self.links_stats.items() if x[1]]
        tab = [tuple(list(x) + self.move_shs.get(x[0], [''])) for x in tab]
        print(tab)
        t = tabulate(tab, headers='', tablefmt='presto', showindex=False)
        m += '\t- %s' % t.replace(
            '\n', '\n\t- '
        ).replace(
            '|', ' :'
        ).replace(
            ' : sh', ' files  -> sh'
        )
        m += '\n\t-> all softwares '
        n = len(scripts)
        m += '(%s screen session' % n
        if n > 1:
            m += 's'
        m += ')  -> sh %s\n' % self.move_sh
        self.link_script = m
        print(m)

    def get_bring_links_scripts(self, links_dir):
        scripts = {}
        chunks = self.get_links_chunks()
        for cdx, (chunk, links) in enumerate(chunks.items()):

            base = '%s/scripts/move%s' % (links_dir, chunk)
            out = '%s_out.txt' % base
            screen_sh = '%s_screen.sh' % base
            sh = '%s.sh' % base

            part, name = '', 'move'
            if chunk[1:].isdigit():
                part += ' [ %s / %s ]' % (chunk[1:], len(chunks))
                name += '_%s_of_%s' % (chunk[1:], len(chunks))
            else:
                part += ' [ %s ]' % chunk[1:]
                name += '_%s' % chunk[1:]
                self.move_shs[chunk[1:]] = ['sh %s' % screen_sh]

            scripts[name] = (screen_sh, out)
            with open(screen_sh, 'w') as screen_o:
                echo = 'Started screen in detached mode (ID: %s)' % name
                echo += '\nCheck whether some moves went wrong: %s' % out
                screen_o.write('screen -dmS %s /bin/bash "%s"\n' % (name, sh))
                screen_o.write('echo "%s"\n' % echo)

            with open(sh, 'w') as o:
                message = 'Bringing data from storage%s' % part
                o.write('touch %s\n' % out)
                o.write('echo "%s"\n' % message)
                o.write('sleep %s\n' % (cdx + 2))
                for link_ in links:
                    dest = link_.replace('${SCRATCH_FOLDER}', '')
                    src = self.links[dest]
                    o.write('if [ -L %s ]; then\n' % dest)
                    o.write("    m0=`md5sum %s | cut -d' ' -f 1`\n" % src)
                    o.write('    rm -rf %s\n' % dest)
                    o.write('    cp -r %s %s\n' % (src, dest))
                    o.write("    m1=`md5sum %s | cut -d' ' -f 1`\n" % dest)
                    o.write('    if [ "$m0" != "$m1" ]; then echo ')
                    o.write('"$m0 $m1 %s %s" >> %s; fi\n' % (src, dest, out))
                    o.write('fi\n\n')
                o.write('echo "done"\n')

        return scripts

    def software_cmds(self):
        m = max(len(x) for x in self.commands.softs) + 1
        for sdx, (name, soft) in enumerate(self.commands.softs.items()):
            self.print_status(m, sdx, name, soft)
            if not len(soft.cmds):
                continue
            self.hash += str(soft.hash)
            self.get_links(soft)
            self.get_modules(name)
            self.get_cmds(soft)
            self.get_main_sh(name, soft)
            if self.cmds:
                self.get_chunks(soft.params['chunks'])
                self.write_jobs(name, soft)
                self.write_main()
                self.write_bash(name, soft)
            self.write_provenance(name, soft)

    def get_links(self, soft):
        self.links_stats[soft.name] = len(soft.links)
        self.soft_links[soft.name] = soft.links
        self.links.update(soft.links)

    def get_sh(self, name: str, chunk: str, soft=None) -> None:
        """

        Parameters
        ----------
        name : str
            Name of the current software of the pipeline workflow
        chunk : str
            Name of the current chunk of commands
        soft
        """
        if soft:
            self.sh = '%s/jobs/run_%s.sh' % (
                soft.dir, chunk.replace('/', '_'))
        else:
            self.sh = '%s/databases/jobs/build_%s_%s.sh' % (
                self.config.dir, name, chunk)
        mkdr(dirname(self.sh))

    def prep_script(self, params: dict) -> None:
        # mandatory options
        self.cmd = [
            'Xhpc',
            '-i', self.sh,
            '-j', self.job_name,
            '-t', str(params['time']),
            '-c', str(params['cpus']),
            '-M', str(params['mem']), params['mem_dim'],
            '--no-stat', '--no-abspath']
        # whether the cpus requests is per node
        if params['nodes']:
            self.cmd.extend(['-n', str(params['nodes'])])
        # always provide an account
        if self.config.account:
            self.cmd.extend(['-a', self.config.account])
        # get the job script file path and use it
        job_script = '%s.slm' % splitext(self.sh)[0]
        if self.config.torque:
            self.cmd.append('--torque')
            job_script = '%s.pbs' % splitext(self.sh)[0]
        self.job_fps.append(job_script)
        self.cmd.extend(['-o', job_script])
        # machine-specific setup: env activating and slurm partition
        if params['machine']:
            self.cmd.append('--%s' % params['machine'])
        if params['partition']:
            self.cmd.append('--p-partition %s' % params['partition'])
        # whether an environment must be used for the current software
        if params['env']:
            self.cmd.extend(['-e', params['env']])
        # setup the scratch location to be used for the current software
        if isinstance(params['scratch'], int):
            self.cmd.extend(['--localscratch', str(params['scratch'])])
        elif params['scratch'] == 'scratch':
            self.cmd.append('--scratch')
        elif params['scratch'] == 'userscratch':
            self.cmd.append('--userscratch')
        if not self.config.verbose:
            self.cmd.append('--quiet')

    def call_cmd(self):
        cmd = ' '.join(self.cmd)
        subprocess.call(cmd.split())
        os.remove(self.sh)

    def write_chunks(self, chunk_keys: list, soft):
        with open(self.sh, 'w') as sh:
            if self.modules:
                sh.write('module purge\n')
            for module in self.modules:
                sh.write('module load %s\n' % module)
            for chunk_key in chunk_keys:
                for cmd in self.cmds[chunk_key]:
                    if soft.params['scratch'] and self.config.jobs:
                        sh.write('%s\n' % cmd)
                    else:
                        sh.write('%s\n' % cmd.replace('${SCRATCH_FOLDER}', ''))

    def get_job_name(self, name: str, chunk: str, soft=None):
        self.job_name = name + '.' + self.pjct
        if soft:
            self.job_name += '.' + ''.join([
                x for x in soft.dir.split('after_')[-1] if x not in 'aeiuoy'])
        self.job_name += '.' + chunk.replace('/', '_')

    def write_dummy_oe(self):
        job_dir = '%s/output' % dirname(self.sh)
        if not isdir(job_dir):
            os.makedirs(job_dir)
        for job_fp in glob.glob('%s/#%s*' % (job_dir, self.job_name)):
            os.remove(job_fp)
        jobs = range(1, random.randrange(2, 4))
        for job in jobs:
            job_fp = 'slurm-%s_000000%s' % (self.job_name, job)
            lines = ['random %s' % r for r in range(5, random.randrange(6, 20))]
            if job == jobs[-1] and random.choice([0, 1]):
                lines.append('Done!')
            for oe in ['.o', '.e']:
                job_fp_oe = '%s/%s%s' % (job_dir, job_fp, oe)
                with open(job_fp_oe, 'w') as o_dev:
                    o_dev.write('\n'.join(lines))

    def write_script(self, soft=None):
        if self.config.jobs:
            if soft:
                self.prep_script(soft.params)
            else:
                self.prep_script(self.config.params)
            self.call_cmd()
        else:
            self.job_fps.append('%s.sh' % splitext(self.sh)[0])

    def write_jobs(self, name: str, soft=None):
        for chunk, chunk_keys in self.chunks.items():
            self.get_sh(name, chunk, soft)
            self.write_chunks(chunk_keys, soft)
            self.get_job_name(name, chunk, soft)
            if self.config.dev:
                self.write_dummy_oe()
            self.write_script(soft)

    def display(self):
        for db_soft, name_main in self.run.items():
            if len(self.run[db_soft]):
                print()
                soft_print = '========== %s scripts ========== ' % db_soft
                print(soft_print)
                self.scripts.append(soft_print)
            for name, main in name_main.items():
                if main.startswith('Input'):
                    main_print = '>%s\n%s' % (name, main)
                else:
                    main_print = '>%s\nsh %s' % (name, main)
                print(main_print)
                self.scripts.append(main_print)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import sys
from os.path import basename, splitext
from metagenomix._io_utils import io_update, to_do, status_update


def viralverify_cmd(
        self,
        contigs: str,
        out: str
) -> str:
    """Collect ViralVerify command.

    Parameters
    ----------
    self : Commands class instance
        .databases : dict
            Path to the reference databases
        .outputs : dict
            All outputs
    contigs : str
        Path to the input contigs fasta file
    out : str
        Path to the output folder

    Returns
    -------
    cmd : str
        ViralVerify command
    """
    cmd = 'export PATH=$PATH:%s' % self.soft.params['path']
    cmd += '\nviralverify'
    cmd += ' -f %s' % contigs
    cmd += ' -o %s' % out
    if self.soft.params['db']:
        cmd += ' --db'
    if self.soft.params['p']:
        cmd += ' --p'
    cmd += ' -t %s' % self.soft.params['cpus']
    cmd += ' --thr %s' % self.soft.params['thr']
    cmd += ' --hmm %s' % '%s/Pfam-A.hmm' % self.databases.paths.get('pfam')
    return cmd


def viralverify(self) -> None:
    """viralVerify classifies contigs (output of metaviralSPAdes or other
    assemblers) as viral, non-viral or uncertain, based on gene content. Also
    for non-viral contigs it can optionally provide plasmid/non-plasmid
    classification.

    viralVerify predicts genes in the contig using Prodigal in the
    metagenomic mode, runs hmmsearch on the predicted proteins and classifies
    the contig as vrial or non-viral by applying the Naive Bayes classifier
    (NBC). For the set of predicted HMMs, viralVerify uses trained NBC to
    classify this set to be viral or chromosomal.

    To improve results in the case of metagenomes with possible host
    contamination, we recommend users to filter out reads that align to the
    host genome prior to assembly. Since viralVerify is based on gene
    classification, it can be used on contigs on any length, and short
    viruses can be detected as long as they contain a recognizable
    virus-specific gene. To help analyze the rapidly growing amount of novel
    data, we have added a script that allows users to construct their own
    training database from a set of viral, chromosomal and plasmid contigs,
    as well as custom HMM database.

    References
    ----------
    Antipov, Dmitry, et al. "Metaviral SPAdes: assembly of viruses from
    metagenomic data." Bioinformatics 36.14 (2020): 4126-4129.

    Notes
    -----
    GitHub  : https://github.com/ablab/viralVerify
    Paper   : https://doi.org/10.1093/bioinformatics/btaa490

    Parameters
    ----------
    self : Commands class instance
        .soft.prev : str
            Previous software in the pipeline
        .dir : str
            Path to pipeline output folder for CoCoNet
        .soft.sam_pool : str
            Sample or co-assembly group name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    if self.soft.prev not in self.config.tools['assembling']:
        sys.exit('[viralVerify] can only be run after assembly (on contigs)')

    pfam = '%s/Pfam-A.hmm' % self.databases.paths.get('pfam')
    if not self.config.dev and to_do(pfam):
        sys.exit('[viralVerify] Needs the Pfam .hmm database in database yaml')

    for (tech, group), inputs in self.inputs[self.sam_pool].items():

        to_dos = status_update(self, tech, [inputs[0]], group=group)

        out = '/'.join([self.dir, tech, self.sam_pool, group])
        self.outputs['dirs'].append(out)

        contigs = inputs[0]
        base = splitext(basename(contigs))[0]
        out_fp = '%s/%s_result_table.csv' % (out, base)
        self.outputs['outs'][(tech, group)] = out_fp

        if self.config.force or to_do(out_fp):
            key = (tech, group)
            cmd = viralverify_cmd(self, contigs, out)
            if to_dos:
                self.outputs['cmds'].setdefault(key, []).append(False)
            else:
                self.outputs['cmds'].setdefault(key, []).append(cmd)
            io_update(self, i_f=contigs, o_d=out, key=key)
            self.soft.add_status(tech, self.sam_pool, 1, group=group)
        else:
            self.soft.add_status(tech, self.sam_pool, 0, group=group)


def threecac(self) -> None:
    """3CAC is a three-class classifier designed to classify contigs in mixed
    metagenome assemblies as phages, plasmids, chromosomes, or uncertain.

    References
    ----------
    Pu, Lianrong, and Ron Shamir. "3CAC: improving the classification of
    phages and plasmids from metagenomic assemblies using assembly graphs."
    bioRxiv (2021).

    Notes
    -----
    GitHub  : https://github.com/Shamir-Lab/3CAC
    Paper   : https://doi.org/10.1101/2021.11.05.467408

    Parameters
    ----------
    self : Commands class instance
        .soft.prev : str
            Previous software in the pipeline
        .dir : str
            Path to pipeline output folder for CoCoNet
        .soft.sam_pool : str
            Sample or co-assembly group name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    print()


def coconet_cmd(
        self,
        fp: str,
        bam: list,
        out_dir: str
) -> str:
    """Collect CoCoNet command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    fp : str
        Path to the contigs file
    bam : list
        Path(s) to spades read mapping BAM files
    out_dir : str
        Path to the output folder

    Returns
    -------
    cmd : str
        CoCoNet command
    """
    cmd = 'coconet run'
    cmd += ' --fasta %s' % fp
    cmd += ' --output %s' % out_dir
    cmd += ' --threads %s' % self.soft.params['cpus']
    for boolean in ['quiet', 'no_rc', 'silent', 'continue',
                    'patience', 'recruit_small_contigs']:
        if self.soft.params[boolean]:
            cmd += ' --%s' % boolean
    for p in [
        'debug', 'flag', 'min_ctg_len', 'min_prevalence', 'min_dtr_size',
        'min_mapping_quality', 'min_aln_coverage', 'fragment_step',
        'n_train', 'n_test', 'batch_size', 'test_batch', 'load_batch',
        'cover_filters', 'cover_kernel', 'cover_stride', 'merge_neurons',
        'kmer', 'wsize', 'wstep', 'n_frags', 'max_neighbors', 'n_clusters',
        'vote_threshold', 'gamma1', 'gamma2', 'fragment_length', 'theta',
        'test_ratio', 'learning_rate'
    ]:
        cmd += ' --%s %s' % (p.replace('_', '-'), self.soft.params[p])
    for p in ['tlen_range', 'compo_neurons', 'cover_neurons']:
        if self.soft.params[p]:
            cmd += ' --%s %s' % (p.replace('_', '-'),
                                 ' '.join(map(str, self.soft.params[p])))
    if bam:
        for b in bam:
            print(kjsrbf)
        cmd += ' --bam %s\n' % ' '.join(bam)
    return cmd


def coconet(self) -> None:
    """CoCoNet (Composition and Coverage Network) is a binning method for viral
    metagenomes. It leverages deep learning to abstract the modeling of the
    k-mer composition and the coverage for binning contigs assembled form
    viral metagenomic data. Specifically, our method uses a neural network to
    learn from the metagenomic data a flexible function for predicting the
    probability that any pair of contigs originated from the same genome.
    These probabilities are subsequently combined to infer bins, or clusters
    representing the species present in the sequenced samples. Our approach
    was specifically designed for diverse viral metagenomes, such as those
    found in environmental samples (e.g., oceans, soil, etc.).

    References
    ----------
    Arisdakessian, Cédric G., et al. "CoCoNet: an efficient deep learning
    tool for viral metagenome binning." Bioinformatics 37.18 (2021): 2803-2810.

    Notes
    -----
    GitHub  : https://github.com/Puumanamana/CoCoNet
    Docs    : https://coconet.readthedocs.io/
    Paper   : https://doi.org/10.1093/bioinformatics/btab213

    Parameters
    ----------
    self : Commands class instance
        .soft.prev : str
            Previous software in the pipeline
        .dir : str
            Path to pipeline output folder for CoCoNet
        .soft.sam_pool : str
            Sample or co-assembly group name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    if self.soft.prev not in self.config.tools['assembling']:
        sys.exit('[coconet] can only be run after assembly')

    bams = {}
    if 'mapping_spades' in self.softs:
        bams = self.softs['mapping_spades'].outputs

    for (tech, group), inputs in self.inputs[self.sam_pool].items():

        to_dos = status_update(self, tech, [inputs[0]], group=group)

        bam = bams.get((tech, group), [])

        out_dir = '/'.join([self.dir, tech, self.sam_pool, group])
        self.outputs['dirs'].append(out_dir)
        self.outputs['outs'][group] = out_dir

        log_fp = '%s/coconet.log' % out_dir
        if self.config.force or to_do(log_fp):
            contigs = inputs[0]
            cmd = coconet_cmd(self, contigs, bam, out_dir)
            key = (tech, group)
            if to_dos:
                self.outputs['cmds'].setdefault(key, []).append(False)
            else:
                self.outputs['cmds'].setdefault(key, []).append(cmd)
            io_update(self, i_f=contigs, o_d=out_dir, key=key)
            self.soft.add_status(tech, self.sam_pool, 1, group=group)
        else:
            self.soft.add_status(tech, self.sam_pool, 0, group=group)


def deepvirfinder(self):
    pass


def deepvirfinder_extractVirs(self):
    pass


def wish(self):
    pass



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import click

from metagenomix import __version__

from metagenomix.scripts.modules import create
from metagenomix.scripts.modules import export
from metagenomix.scripts.modules import monitor
from metagenomix.scripts.modules import merge
from metagenomix.scripts.modules import manage


@click.group(help="Metagenomix command line manager")
@click.version_option(__version__, prog_name="metagenomix")
def standalone_metagenomix():
    pass


standalone_metagenomix.add_command(create.create)
standalone_metagenomix.add_command(export.export)
standalone_metagenomix.add_command(monitor.monitor)
standalone_metagenomix.add_command(manage.manage)
standalone_metagenomix.add_command(merge.merge)


if __name__ == '__main__':
    standalone_metagenomix()



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import re
import socket
import subprocess
import datetime as dt
from os.path import (
    abspath, basename, dirname, expanduser, expandvars, isdir, isfile, splitext)

from metagenomix.core.output import Softwares


def exporter(**kwargs):
    """Prepare an archive for specific pipeline outputs.

    Parameters
    ----------
    kwargs : dict
        All arguments passed in command line, including defaults
    """
    print('\n>>> `metagenomix export` started >>>\n')
    kwargs['command'] = 'export'
    exporting = Exported(**kwargs)
    exporting.get_folder_exp()
    exporting.get_extensions()
    exporting.get_inputs()
    exporting.get_output()
    print('\t>', exporting.out)
    exporting.get_commands()
    exporting.write_job()
    exporting.showdown()
    print('\n<<< `metagenomix export` completed <<<\n')


class Exported(object):

    def __init__(self, **kwargs) -> None:
        self.__dict__.update(kwargs)
        self.softs = Softwares(**kwargs)
        self.time = dt.datetime.now().strftime("%d-%m-%Y_%H-%M-%S")
        self.dir = abspath(self.dir)
        self.export_dir = abspath('%s/_exports' % self.dir)
        self.out = ''
        self.extensions = []
        self.to_exports = []
        self.commands = []

    def get_folder_exp(self):
        self.out = '%s/%s' % (self.export_dir, self.time)
        if not self.local and self.location in os.environ:
            self.out = '%s/exports_%s' % (os.environ[self.location], self.time)

    def get_extensions(self):
        if self.exts:
            print('* Getting extensions to target specific files')
            self.exts = ['.%s' % x if x[0] != '.' else x for x in self.exts]

    def get_inputs(self):
        print('\n* Getting files to export:')
        if self.regex:
            regex = re.compile(
                r'%s' % '|'.join(list(self.regex)), flags=re.IGNORECASE)

        m = ''
        for root, dirs, files in os.walk(self.dir):
            if root == self.dir:
                continue
            soft = root.split('%s/' % self.dir)[-1].split('/')[0]
            if self.softs.names and soft not in self.softs.names:
                continue

            for fil in files:
                if self.regex:
                    if re.search(regex, fil):
                        self.to_exports.append('%s/%s' % (root, fil))
                        continue
                ext = splitext(fil)[1]
                if self.exts:
                    if ext in self.exts or ext[1:] in self.exts:
                        m = '(with extensions "%s" )' % '", "'.join(self.exts)
                        self.to_exports.append('%s/%s' % (root, fil))
                else:
                    self.to_exports.append('%s/%s' % (root, fil))

        print('\t%s files %swill be exported' % (len(self.to_exports), m))

    def get_output(self):
        print('* Getting output archive path')
        self.output = abspath(self.output)
        if not self.local and self.location in os.environ:
            self.output = '%s/%s' % (os.environ[self.location],
                                     basename(self.output))
        self.output = '%s.tar.gz' % splitext(expandvars(self.output))[0]

    def get_commands(self):
        print('* Getting copying commmands')
        self.get_copy_commands()
        print('* Getting archiving commmands')
        self.get_archiving_commands()

    def get_copy_commands(self):
        for to_export in self.to_exports:
            exp = to_export.replace(self.output.rstrip('/'), self.out)
            if isfile(exp):
                continue
            if not isdir(dirname(exp)):
                self.commands.append('mkdir -p %s' % dirname(exp))
            self.commands.append('cp %s %s' % (to_export, exp))

    def get_archiving_commands(self):
        tar_cmd = 'tar czf %s -C %s .' % (self.output, self.out)
        rm_cmd = 'rm -rf %s' % self.out
        self.commands.append(tar_cmd)
        self.commands.append(rm_cmd)

    def write_sh(self):
        job_sh = '%s/jobs/export_%s.sh' % (self.export_dir, self.time)
        if not isdir(dirname(job_sh)):
            os.makedirs(dirname(job_sh))
        with open(job_sh, 'w') as o:
            for command in self.commands:
                o.write('%s\n' % command)
        return job_sh

    def run_xhpc(self, sh):
        if self.torque:
            hpc = '%s.pbs' % splitext(sh)[0]
            pbs = ' --torque'
        else:
            hpc = '%s.slm' % splitext(sh)[0]
            pbs = ''
        cmd = 'Xhpc -i %s -o %s -j xpt_%s -t 2%s' % (sh, hpc, self.time, pbs)
        if self.account:
            cmd += ' -a %s' % self.account
        cmd += ' --quiet'
        subprocess.call(cmd.split())
        os.remove(sh)
        return hpc

    def write_job(self):
        sh = self.write_sh()
        if self.jobs:
            hpc = self.run_xhpc(sh)
            with open(sh, 'w') as o:
                o.write('mkdir %s/jobs/output\n' % self.export_dir)
                o.write('cd %s/jobs/output\n' % self.export_dir)
                if self.torque:
                    o.write('squeue %s\n' % hpc)
                else:
                    o.write('sbatch %s\n' % hpc)
        print('\n* To export, run the following:\nsh %s\n' % sh)

    def showdown(self):
        user = expanduser('~').split('/')[-1]
        hostname = socket.gethostname()
        print('* Then, copy(-edit)-paste to download from server to local:')
        print('scp %s@%s:%s .' % (user, hostname, self.output))



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import click

from metagenomix.exporter import exporter
from metagenomix import __version__


@click.command()
@click.option("-i", "--folder", required=True,
              help='Path to pipeline output folder (`-o` for "create" module)')
@click.option("-o", "--output", required=True, help="Output archive file")
@click.option("-p", "--pipeline",
              help="Path to the file containing the softwares to run in order")
@click.option("-s", "--software", multiple=True,
              help="Software(s) to manage (or all in `-i/-p`)")
@click.option("-e", "--extension", multiple=True, default=None,
              show_default=True, help="Extension to select files")
@click.option("-r", "--regex", multiple=True, default=None,
              show_default=False, help="Regex for file names to select")
@click.option("-l", "--location", default='USERWORK',
              help='If not creating the tar locally, create it there')
@click.option("--local/--no-local", default=False,
              help="Creates the tar locally, and not in the location of `-l`")
@click.option("-a", "--account", show_default=False, default=None,
              help="User account for your HPC (in use for Slurm)")
@click.option("--jobs/--no-jobs", default=True, show_default=True,
              help="Whether to make the archive in a job")
@click.option("--torque/--no-torque", default=False, show_default=True,
              help="Whether to prepare Torque jobs instead of Slurm")
@click.option("--verbose/--no-verbose", default=True, show_default=True,
              help="Whether to show input/outputs and other details")
@click.version_option(__version__, prog_name="metagenomix")
def export(
        folder,
        output,
        pipeline,
        software,
        extension,
        regex,
        location,
        local,
        account,
        jobs,
        torque,
        verbose
):
    """Prepare an archive for specific pipeline outputs."""
    exporter(
        dir=folder,
        output=output,
        pipeline=pipeline,
        softwares=software,
        exts=extension,
        regex=regex,
        location=location,
        local=local,
        account=account,
        jobs=jobs,
        torque=torque,
        verbose=verbose
    )



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import click

from metagenomix.monitoring import monitoring
from metagenomix import __version__


@click.command()
@click.option(
    "-i", "--fastq-illumina", multiple=True,
    help="Path to short Illumina reads fastq files folder(s)")
@click.option(
    "-j", "--fastq-pacbio", multiple=True,
    help="Path to long PacBio reads fastq files folder(s)")
@click.option(
    "-k", "--fastq-nanopore", multiple=True,
    help="Path to long MinION Nanopore reads fastq files folder")
@click.option(
    "-o", "--out-dir", required=True, help="Path to pipeline output folder")
@click.option(
    "-s", "--summary", help="Path to summary output file")
@click.option(
    "-m", "--metadata", required=True, help="Path to the metadata file")
@click.option(
    "-d", "--databases", show_default=True, help="Databases (yaml file)")
@click.option(
    "-p", "--pipeline", show_default=True, help="Softwares to pipeline")
@click.option(
    "-u", "--user-params", show_default=True, help="Server run parameters")
@click.option(
    "-c", "--co-assembly", show_default=True,
    help="Metadata column(s) defining the co-assembly groups (yaml file)")
@click.option(
    "-t", "--strains", show_default=True,
    help="Species for strain-level analyses (yaml file)")
@click.option(
    "--show-params/--no-show-params", default=False, show_default=False,
    help="Show all possible parameters for all softwares of your pipeline")
@click.option(
    "--force/--no-force", default=False, show_default=True,
    help="Check as if the re-writing of scripts for all commands was planned")
@click.option(
    "--verbose/--no-verbose", default=False, show_default=True,
    help="Whether to show input/outputs and other details")
@click.option(
    "--dev/--no-dev", default=False, show_default=True,
    help="For development...")
@click.version_option(__version__, prog_name="metagenomix")
def monitor(
        metadata,
        fastq_illumina,
        fastq_pacbio,
        fastq_nanopore,
        out_dir,
        summary,
        co_assembly,
        databases,
        pipeline,
        user_params,
        strains,
        show_params,
        force,
        verbose,
        dev
):
    """Check IO/job status of your pipeline configuration."""
    monitoring(
        meta_fp=metadata,
        illumina_dirs=fastq_illumina,
        pacbio_dirs=fastq_pacbio,
        nanopore_dirs=fastq_nanopore,
        output_dir=out_dir,
        summary_fp=summary,
        coassembly_yml=co_assembly,
        databases_yml=databases,
        pipeline_tsv=pipeline,
        user_params_yml=user_params,
        strains_yml=strains,
        show_params=show_params,
        force=force,
        verbose=verbose,
        dev=dev
    )
